#include <string.h>
#include "msClassic.h"
#include "printf.h"
#include "msioLL.h"
#include "msioHL.h"
#include "timer.h"
#include "util.h"

//MS regs
#define MS_REG_INT					0x01	//RO
#define MS_REG_STA0					0x02	//RO
#define MS_REG_STA1					0x03	//RO
#define MS_REG_TYPE					0x04	//RO
#define MS_REG_IFACE				0x05	//RO
#define MS_REG_CATEGORY				0x06	//RO
#define MS_REG_CLASS				0x07	//RO

#define MS_REG_UNK_HI				0x09	//RO	io related. length of io attributes?
#define MS_REG_UNK_LO				0x0A	//RO	io related. length of io attributes?


#define MS_REG_SYSPARAM				0x10
#define MS_REG_BLK_ADR_HI			0x11
#define MS_REG_BLK_ADR_MID			0x12
#define MS_REG_BLK_ADR_LO			0x13
#define MS_REG_ACCESSTYP			0x14	//0x00 - block, 0x20 - page, 0x40 - OOB only, 0x80 - overwrite
#define MS_REG_PG_ADDR				0x15
#define MS_REG_OVERWRI_FLG			0x16	//struct MsOobData:
#define MS_REG_MANAGE_FLG			0x17
#define MS_REG_LOGICAL_ADR_HI		0x18
#define MS_REG_LOGICAL_ADR_LO		0x19
///6 more bytes here that are unused but are stored per block with ECC
#define MS_REG_READ_FIRST			MS_REG_INT
#define MS_REG_READ_LAST			MS_REG_LOGICAL_ADR_LO
#define MS_REG_WRITE_FIRST			MS_REG_SYSPARAM
#define MS_REG_WRITE_LAST			MS_REG_LOGICAL_ADR_LO
#define MS_NUM_REGS_READ			(MS_REG_READ_LAST + 1 - MS_REG_READ_FIRST)
#define MS_NUM_REGS_WRITE			(MS_REG_WRITE_LAST + 1 - MS_REG_WRITE_FIRST)

//sta1 bits (some are ORRed together as individual use is not clear
#define MS_STA1_UNCORRECTABLE		0x15
#define MS_STA1_CORRECTED			0x2a

#define MS_MG_FLG_SYSTEM			0x04	//system-used page (not user data)

//commands
#define MS_CMD_RESET				0x3c
#define MS_CMD_PAGE_READ			0xaa

#define MS_PRO_CMD_FORMAT			0x10
#define MS_PRO_CMD_SLEEP			0x11
#define MS_PRO_CMD_READ				0x20
#define MS_PRO_CMD_WRITE			0x21
#define MS_PRO_CMD_ATTRIB_READ		0x24
#define MS_PRO_CMD_STOP				0x25
#define MS_PRO_CMD_ERASE			0x26

enum MsReadResult {
	MsReadOK,
	MsReadCorrected,
	MsReadUncorrectable,
	MsReadError,
};

struct MsOobData {
	uint8_t overwriFlg;
	uint8_t manageFlag;
	uint16_t logicalAddr;	//BE on media
};






bool msClassicIsMsClassic(const uint8_t *firstEightRegs)
{
	return firstEightRegs[MS_REG_TYPE] == 0xff;
}

static bool msPrvCmd(uint_fast8_t cmd, uint_fast8_t expectedIntVal)
{
	uint8_t rx, intValOut;
	
	if (!msioHlCmd(cmd)) {
		pr("cmd broken\n");
		return false;
	}
	
	if (!msioHlCmdEndWaitInt(TIMER_TICKS_PER_SEC * 15 / 10, &intValOut)) {
		pr("comand timed out\n");
		return false;
	}
	
	if (intValOut != expectedIntVal) {
		
		pr("command int was %02xh but expecvted was %02xh\n", intValOut, expectedIntVal);
		return false;
	}
	
	return true;
}

static enum MsReadResult msPrvReadPage(uint32_t block, uint8_t page, uint8_t *data, struct MsOobData *oob)
{
	union {	//to save stack
		uint8_t wrRegs[MS_NUM_REGS_WRITE];
		uint8_t rdRegs[MS_NUM_REGS_READ];
		
	} regs = {
			.wrRegs = {
				[MS_REG_BLK_ADR_HI - MS_REG_WRITE_FIRST] = block >> 16,
				[MS_REG_BLK_ADR_MID - MS_REG_WRITE_FIRST] = block >> 8,
				[MS_REG_BLK_ADR_LO - MS_REG_WRITE_FIRST] = block >> 0,
				[MS_REG_ACCESSTYP - MS_REG_WRITE_FIRST] = data ? 0x20 : 0x40,
				[MS_REG_PG_ADDR - MS_REG_WRITE_FIRST] = page,
			},
		};
	
	//set regs for the read
	
	if (!msioHlRegsWrite(regs.wrRegs, MS_NUM_REGS_WRITE)) {
		pr("failed to write regs\n");
		return MsReadError;
	}
	
	if (!msPrvCmd(MS_CMD_PAGE_READ, 0xa0)) {
		pr("failed to initiate the command\n");
		return MsReadError;
	}
	
	//we need the status regs in all cases to see correctability info, so always read the regs
	if (!msioHlRegsRead(regs.rdRegs, MS_NUM_REGS_READ)) {
		pr("get REGS fail\n");
		return MsReadError;
	}
	pr("REGS:\n");
	showBuf(regs.rdRegs, MS_NUM_REGS_READ);
	
	//if OOB was requested, provide it
	if (oob) {
		
		oob->overwriFlg = regs.rdRegs[MS_REG_OVERWRI_FLG - MS_REG_READ_FIRST];
		oob->manageFlag = regs.rdRegs[MS_REG_MANAGE_FLG - MS_REG_READ_FIRST];
		oob->logicalAddr = readBE16(regs.rdRegs + MS_REG_LOGICAL_ADR_HI - MS_REG_READ_FIRST);
	}
	
	//if data was requested, read that
	if (data) {
		
		if (!msioLlTpcR(MS_RD_LDATA, data, 512)) {
			pr("get FATA fail\n");
			return MsReadError;
		}
	}
	
	//sort out correctability
	if (regs.rdRegs[MS_REG_STA1 - MS_REG_READ_FIRST] & MS_STA1_UNCORRECTABLE)
		return MsReadUncorrectable;
	
	if (regs.rdRegs[MS_REG_STA1 - MS_REG_READ_FIRST] & MS_STA1_CORRECTED)
		return MsReadCorrected;
	
	return MsReadOK;
}

void msClassicDemo(void)
{
	uint8_t buf[512];
	struct MsOobData oob;
	enum MsReadResult rr;
	uint_fast8_t block;
	
	
	if (!msioHlSetRegAddrs(MS_REG_READ_FIRST, MS_NUM_REGS_READ, MS_REG_WRITE_FIRST, MS_NUM_REGS_WRITE)) {
		pr("failed to set reg addrs\n");
		return;
	}
	
	
	//read info page. not so easy for MS since we have to do software wear leveling, and we do not even know where to look
	//we look in the first 17 blocks, same as the Clie driver does
	for (block = 0; block < 17; block++) {
		
		rr = msPrvReadPage(block, 0, buf, &oob);
		if (rr == MsReadError) {
			pr("read page error\n");
			return;
		}
		
		pr("read %u.0: ret = %u, OOB = {ov=%02xh, mng=%02xh, la=%04xh}\n", block, rr, oob.overwriFlg, oob.manageFlag, oob.logicalAddr);
		showBuf(buf, 512);

		if (rr == MsReadUncorrectable) {
			pr("%u.0 is uncorrectable\n", block);
			continue;
		}
		
		if (oob.manageFlag & MS_MG_FLG_SYSTEM) {
			pr("%u.0 has unexpected flags for info page\n", block);
			continue;
		}
		
		if (buf[0] != 0x00 || buf[1] != 0x01) {
			pr("%u.0 has unexpected first bytes for info page\n", block);
			continue;
		}
		
		pr("possible boot block flagged\n");
	}
}

